﻿Imports Microsoft.VisualBasic
Imports System
Imports System.Collections.Generic
Imports System.Linq
Imports System.ServiceModel
Imports System.Threading
Imports PeerWcfComm

Namespace WorkerPeerApplication
    ''' <summary>
    ''' Defintion of Worker peer. It participates in IWorker and IWorkBackup communication.
    ''' </summary>
    <ServiceBehavior(InstanceContextMode:=InstanceContextMode.Single)> _
    Public Class Worker
        Implements IWorker, IWorkBackup
#Region "Private Fields..."

        Private _host As ServiceHost

        Private _workerFactory As ChannelFactory(Of IWorker)
        Private _spectatorFactory As ChannelFactory(Of ISpectator)
        Private _backupFactory As ChannelFactory(Of IWorkBackup)
        Private _workerClientChannel As IWorker
        Private _spectatorClientChannel As ISpectator
        Private _backupClientChannel As IWorkBackup

        Private _id As Guid
        Private _state As WorkerState
        Private _currentWorkItem As WorkItem

        Private _workQueue As Queue(Of WorkItem)
        Private _workers As Dictionary(Of Guid, WorkerStatus)
        Private _workItemBackups As Dictionary(Of Guid, WorkItemBackupRequest)

        Private _heartbeat As Timer
        Private _backupMonitor As Timer

#End Region

#Region "Public Ctors and Methods..."
        Public Sub New()
            _id = Guid.NewGuid()
            _state = WorkerState.Offline
            _workQueue = New Queue(Of WorkItem)()
            _workers = New Dictionary(Of Guid, WorkerStatus)()
            _workItemBackups = New Dictionary(Of Guid, WorkItemBackupRequest)()
        End Sub

        Public Sub Start()
            _host = New ServiceHost(Me)
            _host.Open()

            _workerFactory = New ChannelFactory(Of IWorker)("IWorkerClient")
            _workerClientChannel = _workerFactory.CreateChannel()

            _spectatorFactory = New ChannelFactory(Of ISpectator)("ISpectatorClient")
            _spectatorClientChannel = _spectatorFactory.CreateChannel()

            _backupFactory = New ChannelFactory(Of IWorkBackup)("IWorkBackupClient")
            _backupClientChannel = _backupFactory.CreateChannel()

            Dim pn As PeerNode = (CType(_workerClientChannel, IClientChannel)).GetProperty(Of PeerNode)()
            pn.MessagePropagationFilter = New RemoteOnlyMessagePropagationFilter()

            ThreadPool.QueueUserWorkItem(AddressOf MonitorWork)

            _state = WorkerState.Ready
            _heartbeat = New Timer(AddressOf Me.SendHeartbeat, Nothing, New TimeSpan(0), New TimeSpan(0, 0, 1)) 'Time should be read in from config.

            _backupMonitor = New Timer(AddressOf Me.MonitorBackups, Nothing, New TimeSpan(0), New TimeSpan(0, 1, 0)) 'Time should be read in from config.
        End Sub

        Public Sub [Stop]()
            _state = WorkerState.Offline
            _heartbeat.Dispose()
            _backupMonitor.Dispose()
            Me.SendHeartbeat(Nothing)
            _workerFactory.Close()
            _spectatorFactory.Close()
            _backupFactory.Close()
            _host.Close()
        End Sub
#End Region

#Region "Private Methods..."
        Private Sub SendHeartbeat(ByVal context As Object)
            Dim ws As New WorkerStatus()
            ws.WorkerId = _id
            ws.State = _state
            ws.PendingWorkItemCount = _workQueue.Count
            ws.CurrentWorkItem = If((_currentWorkItem Is Nothing), Guid.Empty, _currentWorkItem.WorkItemId)

            _workerClientChannel.Heartbeat(ws)
        End Sub

        Private Sub MonitorWork(ByVal context As Object)
            If _workQueue.Count > 0 Then
                SyncLock _workQueue
                    _currentWorkItem = _workQueue.Dequeue()
                End SyncLock

                DoWorkInternal(_currentWorkItem)
                _currentWorkItem = Nothing
            End If

            If _state <> WorkerState.Offline Then
                If _workQueue.Count > 1 Then
                    Dim firstIdleWorker As WorkerStatus = GetNextIdleWorker()

                    If firstIdleWorker IsNot Nothing Then
                        Dim workToForward As WorkItem

                        SyncLock _workQueue
                            workToForward = _workQueue.Dequeue()
                        End SyncLock

                        _workerClientChannel.DoWork(firstIdleWorker.WorkerId, workToForward)
                    End If
                End If

                Thread.Sleep(100) 'Brief pause to not spin up CPU cycles!
                ThreadPool.QueueUserWorkItem(AddressOf MonitorWork)
            End If
        End Sub

        Private Sub MonitorBackups(ByVal context As Object)
            Dim backups As New List(Of WorkItemBackupRequest)()
            Dim timeOutThreshold As New TimeSpan(0, 1, 0) 'This should be read in from config.

            SyncLock _workItemBackups
                backups.AddRange(_workItemBackups.Values)

                For Each backup As WorkItemBackupRequest In backups
                    'If we've passed the RestoreAt time, we need to
                    'take on this work item.
                    If backup.RestoreAt < DateTime.Now Then
                        SyncLock _workQueue
                            _workQueue.Enqueue(backup.Item)
                        End SyncLock

                        _workItemBackups.Remove(backup.Item.WorkItemId)

                        Console.WriteLine("Restored item '{0}'", backup.Item.WorkItemId)

                        'This work item has not been deleted yet. This is suspicious
                        'so investigate further.
                    ElseIf backup.BackupRequestTimeStamp.Add(timeOutThreshold) > DateTime.Now Then
                        'Check to see if the original worker is still operating.
                        'If so, assume that either:
                        '  1. It's busy and just hasn't gotten to this yet.
                        '  2. It did process this but somehow didn't tell us
                        '      to delete the backup.
                        'Regardless, we won't do anything and the backup
                        'remains in our cache.
                        If (Not WorkerIsStillActive(backup.OriginalWorkerId)) Then
                            'The machine doesn't seem to be actively working.
                            'Set a deadline for restoration of the work item.
                            backup.RestoreAt = DateTime.Now.Add(timeOutThreshold)

                            'Tell the world our intentions to restore this.
                            _backupClientChannel.RestoringBackup(Me._id, backup.Item.WorkItemId)

                            Console.WriteLine("Planning to restore '{0}'", backup.Item.WorkItemId)
                        End If
                    End If
                Next backup
            End SyncLock
        End Sub

        Private Function WorkerIsStillActive(ByVal workerId As Guid) As Boolean
            Dim active As Boolean = False

            SyncLock _workers
                If _workers.ContainsKey(workerId) Then
                    Dim sts As WorkerStatus = _workers(workerId)
                    If sts.State = WorkerState.Busy Then
                        active = True
                    ElseIf sts.State = WorkerState.Ready AndAlso sts.PendingWorkItemCount > 0 Then
                        active = True
                    End If
                End If
            End SyncLock

            Return active
        End Function

        Private Function GetNextIdleWorker() As WorkerStatus
            Dim workers As New List(Of WorkerStatus)()

            SyncLock _workers
                workers.AddRange(_workers.Values)
            End SyncLock

            Return workers.Where(Function(w) (w.State = WorkerState.Ready AndAlso w.PendingWorkItemCount = 0)).FirstOrDefault()
        End Function

        Private Sub DoWorkInternal(ByVal wrk As WorkItem)
            Me._state = WorkerState.Busy
            _spectatorClientChannel.WorkItemStateChanged(New WorkItemStatus() With {.WorkerId = _id, .WorkItemId = wrk.WorkItemId, .NewState = WorkItemState.Executing})

            'Doing our work...
            Thread.Sleep(New TimeSpan(0, 0, wrk.PretendToWorkSeconds))

            Me._backupClientChannel.DeleteBackup(wrk.WorkItemId)

            Me._state = WorkerState.Ready
            _spectatorClientChannel.WorkItemStateChanged(New WorkItemStatus() With {.WorkerId = _id, .WorkItemId = wrk.WorkItemId, .NewState = WorkItemState.Finished})

        End Sub
#End Region

#Region "IWorker Members"

        Private Sub Heartbeat(ByVal status As WorkerStatus) Implements IWorker.Heartbeat
            SyncLock _workers
                _workers(status.WorkerId) = status
            End SyncLock
        End Sub

        Private Sub DoWork(ByVal workerId As Guid, ByVal work As WorkItem) Implements IWorker.DoWork
            If workerId = _id Then
                SyncLock _workQueue
                    _workQueue.Enqueue(work)
                End SyncLock

                Me._backupClientChannel.BackupItem(New WorkItemBackupRequest() With {.Item = work, .OriginalWorkerId = _id})
            End If
        End Sub

#End Region

#Region "IWorkBackup Members"

        Private Sub BackupItem(ByVal request As WorkItemBackupRequest) Implements IWorkBackup.BackupItem
            request.BackupRequestTimeStamp = DateTime.Now
            request.RestoreAt = DateTime.MaxValue

            SyncLock _workItemBackups
                _workItemBackups.Add(request.Item.WorkItemId, request)
            End SyncLock
        End Sub

        Private Sub DeleteBackup(ByVal workItemId As Guid) Implements IWorkBackup.DeleteBackup
            SyncLock _workItemBackups
                If _workItemBackups.ContainsKey(workItemId) Then
                    _workItemBackups.Remove(workItemId)
                End If
            End SyncLock
        End Sub

        Private Sub RestoringBackup(ByVal workerId As Guid, ByVal workItemId As Guid) Implements IWorkBackup.RestoringBackup
            SyncLock _workItemBackups
                'Only proceed if we have this item backed up.
                If _workItemBackups.ContainsKey(workItemId) Then
                    'Get the backup request for this work item
                    Dim req As WorkItemBackupRequest = _workItemBackups(workItemId)

                    'Check to see if we've decided to restore this.
                    If req.RestoreAt <> DateTime.MaxValue Then
                        'The restorer with the greatest id wins!
                        If Me._id.CompareTo(workItemId) < 0 Then
                            _workItemBackups.Remove(workItemId)
                        End If
                    Else
                        'We haven't decided to restore it and someone else wants to.
                        'So let them.
                        _workItemBackups.Remove(workItemId)
                    End If
                End If
            End SyncLock
        End Sub

#End Region
    End Class
End Namespace
